# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import ActionWorkflowCategoryClass
from .....schema_classes import ActionWorkflowEntrypointClass
from .....schema_classes import ActionWorkflowEntrypointTypeClass
from .....schema_classes import ActionWorkflowFieldClass
from .....schema_classes import ActionWorkflowFieldConditionClass
from .....schema_classes import ActionWorkflowFieldConditionTypeClass
from .....schema_classes import ActionWorkflowFormClass
from .....schema_classes import ActionWorkflowFormRequestClass
from .....schema_classes import ActionWorkflowFormRequestFieldClass
from .....schema_classes import ActionWorkflowInfoClass
from .....schema_classes import ActionWorkflowRequestAccessClass
from .....schema_classes import ActionWorkflowRequestStepStateClass
from .....schema_classes import ActionWorkflowSingleFieldValueConditionClass
from .....schema_classes import ActionWorkflowStepClass
from .....schema_classes import ActionWorkflowStepActorsClass
from .....schema_classes import ActionWorkflowStepDynamicAssignmentClass
from .....schema_classes import ActionWorkflowStepDynamicAssignmentTypeClass
from .....schema_classes import ActionWorkflowStepTypeClass
from .....schema_classes import ActionWorkflowTriggerClass
from .....schema_classes import ActionWorkflowTriggerTypeClass


ActionWorkflowCategory = ActionWorkflowCategoryClass
ActionWorkflowEntrypoint = ActionWorkflowEntrypointClass
ActionWorkflowEntrypointType = ActionWorkflowEntrypointTypeClass
ActionWorkflowField = ActionWorkflowFieldClass
ActionWorkflowFieldCondition = ActionWorkflowFieldConditionClass
ActionWorkflowFieldConditionType = ActionWorkflowFieldConditionTypeClass
ActionWorkflowForm = ActionWorkflowFormClass
ActionWorkflowFormRequest = ActionWorkflowFormRequestClass
ActionWorkflowFormRequestField = ActionWorkflowFormRequestFieldClass
ActionWorkflowInfo = ActionWorkflowInfoClass
ActionWorkflowRequestAccess = ActionWorkflowRequestAccessClass
ActionWorkflowRequestStepState = ActionWorkflowRequestStepStateClass
ActionWorkflowSingleFieldValueCondition = ActionWorkflowSingleFieldValueConditionClass
ActionWorkflowStep = ActionWorkflowStepClass
ActionWorkflowStepActors = ActionWorkflowStepActorsClass
ActionWorkflowStepDynamicAssignment = ActionWorkflowStepDynamicAssignmentClass
ActionWorkflowStepDynamicAssignmentType = ActionWorkflowStepDynamicAssignmentTypeClass
ActionWorkflowStepType = ActionWorkflowStepTypeClass
ActionWorkflowTrigger = ActionWorkflowTriggerClass
ActionWorkflowTriggerType = ActionWorkflowTriggerTypeClass

# fmt: on
