{
  "type": "record",
  "Aspect": {
    "name": "datasetDeprecation"
  },
  "Deprecated": true,
  "name": "DatasetDeprecation",
  "namespace": "com.linkedin.pegasus2avro.dataset",
  "fields": [
    {
      "Searchable": {
        "fieldType": "BOOLEAN",
        "weightsPerFieldValue": {
          "true": 0.5
        }
      },
      "type": "boolean",
      "name": "deprecated",
      "doc": "Whether the dataset is deprecated by owner."
    },
    {
      "type": [
        "null",
        "long"
      ],
      "name": "decommissionTime",
      "default": null,
      "doc": "The time user plan to decommission this dataset."
    },
    {
      "type": "string",
      "name": "note",
      "doc": "Additional information about the dataset deprecation plan, such as the wiki, doc, RB."
    },
    {
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      },
      "type": [
        "null",
        "string"
      ],
      "name": "actor",
      "default": null,
      "doc": "The corpuser URN which will be credited for modifying this deprecation content.",
      "Urn": "Urn"
    }
  ],
  "doc": "Dataset deprecation status\nDeprecated! This aspect is deprecated in favor of the more-general-purpose 'Deprecation' aspect."
}